# coding: utf-8
from __future__ import unicode_literals, print_function, division

from clldutils.testing import WithTempDir
from clldutils.path import Path


class TestCase(WithTempDir):
    def setUp(self):
        super(TestCase, self).setUp()

    def data_path(self, *comps):
        return Path(__file__).parent.joinpath('data', *comps)



# coding: utf-8
from __future__ import unicode_literals, print_function, division

from pyclpa.tests.util import TestCase
from pyclpa.wordlist import Wordlist


class Tests(TestCase):
    def _make_one(self, path=None):
        return Wordlist.from_file(path or self.data_path('KSL.tsv'))

    def test_load_wordlist(self):
        wl = self._make_one()
        self.assertIn('TOKENS', wl[0])
        self.assertEqual(len(wl), 1400)

    def test_write_wordlist(self):
        wl = self._make_one()
        out = self.tmp_path('xxx')
        wl.write(out)
        wl2 = self._make_one(out)
        self.assertEqual(wl, wl2)

    def test_check_wordlist(self):
        wl = self._make_one()
        sounds, errors = wl.check()
        assert errors['convertable'] == 13
        assert errors['non-convertable'] == 4
        assert sounds['t']['frequency'] == 223

        sounds, errors = wl.check(rules=self.data_path('KSL.rules'))
        assert errors['non-convertable'] == 3



# coding: utf-8
from __future__ import unicode_literals, print_function, division
from unittest import TestCase

from pyclpa.base import get_clpa, CLPA


class Tests(TestCase):
    def setUp(self):
        self.clpa = get_clpa()
        self.clpa2 = CLPA(rules=dict(th='t'))

    def test_check_sequence(self):
        test1 = self.clpa.check_sequence(['t', 'e', 's', 't'])
        assert '?' not in test1[0]
        test2 = self.clpa.check_sequence('t e s t')
        assert '?' not in test2[0]
        test3 = self.clpa.check_sequence('t e th s t')
        assert test3[0][2][1] == 'ʰ'
        test4 = self.clpa.check_sequence('t X s t')
        assert test4[0][1] == '?'
        test5 = self.clpa.check_sequence('ˈt e s t')
        assert test5[0][0][1] == 't'
        test6 = self.clpa2.check_sequence('th e')
        assert test6[0][0] == 't'

    def test_segment2clpa(self):
        assert self.clpa.segment2clpa('t') == 'c118'
        assert self.clpa.segment2clpa("'t") == 'c118'



# coding: utf-8
from __future__ import unicode_literals, print_function, division

from pyclpa.tests.util import TestCase


class Tests(TestCase):
    def test_local_path(self):
        from pyclpa.util import local_path

        assert local_path('bla').name == 'bla'

    def test_write_CLPA(self):
        from pyclpa.util import write_CLPA, load_CLPA

        write_CLPA(load_CLPA(), self.tmp_path('bla'))

    def test_load_whitelist(self):
        from pyclpa.util import load_whitelist

        assert load_whitelist()['t']['ID'] == 'c118'

    def test_load_alias(self):
        from pyclpa.util import local_path, load_alias

        assert load_alias(local_path('alias.tsv'))['ɡ'] == 'g'

    def test_check_string(self):
        from pyclpa.util import check_string, load_whitelist

        check = check_string('m a tt i s', load_whitelist())
        assert check[2] == '?'

    def test_find_token(self):
        from pyclpa.util import find_token, load_whitelist, load_alias

        wl = load_whitelist()
        patterns = load_alias('patterns.tsv')
        assert not find_token('t', {}, {}, {}, {}, [])
        assert find_token('t', wl, {}, {}, {}, []) == 't'
        assert find_token('th', wl, {'h': 'ʰ'}, {}, {}, []) == 'tʰ'
        assert find_token('th', wl, {}, {'th': 'x'}, {}, []) == 'x'
        with self.assertRaises(ValueError):
            find_token('th', wl, {}, {'th': 'X'}, {}, [])
        assert find_token('th', wl, {}, {}, patterns, []) == 'tʰ'
        assert find_token('th', wl, {}, {}, {}, ['h']) == 't'



# coding: utf-8
"""
Main command line interface to the pyclpa package.
"""
from __future__ import unicode_literals, print_function
import sys
from collections import defaultdict, OrderedDict

from clldutils.clilib import ArgumentParser, ParserError
from clldutils.path import Path
from clldutils.dsv import UnicodeWriter

from pyclpa.util import check_string, load_whitelist
from pyclpa.wordlist import Wordlist


def report(args):
    """
    clpa report <FILE> [rules=FILE] [format=md|csv|cldf] [outfile=FILENAME]

    Note
    ----

    * Rules point to a tab-separated value file in which source and target are
      given to convert a segment to another segment to be applied on a
      data-set-specific basis which may vary from dataset to dataset and can thus
      not be included as standard clpa behaviour.
    * Input file needs to be in csv-format, with tabstop as separator, and it
      needs to contain one column named "TOKENS".
    * format now allows for md (MarkDown), csv (CSV, tab as separator), or cldf
      (no pure cldf but rather current lingpy-csv-format). CLDF format means
      that the original file will be given another two columns, one called
      CLPA_TOKENS, one called CLPA_IDS.
    * if you specify an outfile from the input, the data will be written to
      file instead showing it on the screen.

    """
    if len(args.args) < 1:
        raise ParserError('not enough arguments')

    # get keywords from arguments @xrotwang: is there any better way to do so?
    settings = defaultdict(str)
    settings['format'] = 'md'
    fname = None
    for arg in args.args:
        if '=' in arg:
            key, val = arg.split('=')
            settings[key] = val
        else:
            fname = arg

    if not fname:
        raise ParserError('no filename passed as argument')

    wordlist = Wordlist.from_file(fname)
    sounds, errors = wordlist.check(rules=settings['rules'])

    if settings['format'] not in ['md', 'csv']:
        text = wordlist.write(settings['outfile'] or None)
        if not settings['outfile']:
            print(text)
        return

    segments = OrderedDict([('existing', []), ('missing', []), ('convertible', [])])
    for k in sorted(
        sounds, key=lambda x: (sounds[x]['frequency'], sounds[x]['id']), reverse=True
    ):
        type_, symbol = None, None
        if k == sounds[k]['clpa']:
            type_, symbol = 'existing', k
        elif sounds[k]['clpa'] == '?':
            type_, symbol = 'missing', k
        else:
            check = sounds[k]['clpa']
            if k != check != '?':
                type_, symbol = 'convertible', k + ' >> ' + sounds[k]['clpa']
        if type_ and symbol:
            segments[type_].append([symbol, sounds[k]['id'], sounds[k]['frequency']])

    if settings['format'] == 'csv':
        with UnicodeWriter(settings['outfile'] or None, delimiter='\t') as writer:
            for key, items in segments.items():
                for i, item in enumerate(items):
                    writer.writerow([i + 1] + item + [key])
        if not settings['outfile']:
            print(writer.read())
        return

    text = []
    header_template = """
# {0} sounds

| number | sound | clpa | frequency |
| ------:| ----- | ---- | ---------:|"""

    for key, items in segments.items():
        text.append(header_template.format(key.capitalize()))
        for i, item in enumerate(items):
            text.append("| {0} | {1[0]} | {1[1]} | {1[2]} |".format(i + 1, item))

    text = '\n'.join(text)
    if settings['outfile']:
        with Path(settings['outfile']).open('w', encoding='utf8') as fp:
            fp.write(text)
    else:
        print(text)


def check(args):
    """
    clpa check <STRING>
    """
    if len(args.args) != 1:
        raise ParserError('only one argument allowed')
    check = check_string(args.args[0], load_whitelist())
    print('\t'.join(args.args[0].split(' ')))
    print('\t'.join(check))


def main():  # pragma: no cover
    parser = ArgumentParser('pyclpa', report, check)
    sys.exit(parser.main())



# coding: utf-8
from __future__ import unicode_literals, print_function, division
import unicodedata
import re

from clldutils.path import Path
from clldutils import jsonlib


def local_path(*comps):
    """Helper function to create a local path to the current directory of CLPA"""
    return Path(__file__).parent.joinpath('data', *comps)


def load_CLPA():
    """
    Load the main data file.
    """
    return jsonlib.load(local_path('clpa.main.json'))


def write_CLPA(clpadata, path):
    """
    Basic function to write clpa-data.
    """
    if isinstance(path, Path):
        outdir, fname = path.parent, path.name
    else:
        outdir, fname = local_path(), path  # pragma: no cover
    old_clpa = load_CLPA()
    jsonlib.dump(old_clpa, outdir.joinpath(fname + '.bak'), indent=4)
    jsonlib.dump(clpadata, outdir.joinpath(fname), indent=4)


def load_whitelist():
    """
    Basic function to load the CLPA whitelist.
    """
    _clpadata = jsonlib.load(local_path('clpa.main.json'))
    whitelist = {}
    for group in ['consonants', 'vowels', 'markers', 'tones', 'diphtongs']:
        for val in _clpadata[group]:
            whitelist[_clpadata[val]['glyph']] = _clpadata[val]
            whitelist[_clpadata[val]['glyph']]["ID"] = val

    return whitelist


def load_alias(_path):
    """
    Alias are one-character sequences which we can convert on a step-by step
    basis by applying them successively to all subsegments of a segment.
    """
    path = Path(_path)
    if not path.is_file():
        path = local_path(_path)

    alias = {}
    with path.open(encoding='utf-8') as handle:
        for line in handle:
            if not line.startswith('#') and line.strip():
                source, target = line.strip().split('\t')
                alias[eval('"' + source + '"')] = eval('r"' + target + '"')
    return alias


def split(string):
    return string.split(' ')


def join(tokens):
    return ' '.join(tokens)


def check_string(seq, whitelist):
    return [
        '*' if t in whitelist else '?' for t in split(unicodedata.normalize('NFC', seq))]


def find_token(token, whitelist, alias, explicit, patterns, delete):
    if token in whitelist:
        return token

    # first run, delete useless stuff
    tokens = list(token)
    for i, t in enumerate(tokens):
        if t in delete:
            tokens[i] = ''
    new_token = ''.join(tokens)
    if new_token in whitelist:
        return new_token

    # third run, explicit match
    if new_token in explicit:
        new_token = explicit[new_token]
        if new_token in whitelist:
            return new_token
        raise ValueError(
            "Explicit list does not point to whitelist with sound «{0}»".format(
                new_token))

    # second run, replace
    tokens = list(new_token)
    for i, t in enumerate(tokens):
        if t in alias:
            tokens[i] = alias[t]

    new_token = ''.join(tokens)
    if new_token in whitelist:
        return new_token

    # forth run, pattern matching
    for source, target in patterns.items():
        search = re.search(source, new_token)
        if search:
            new_token = re.sub(source, target, new_token)
        if new_token in whitelist:
            return new_token
    return False



# coding: utf8
from __future__ import unicode_literals, print_function, division
import unicodedata
from collections import Counter

from clldutils.path import Path
from clldutils.dsv import reader, UnicodeWriter

from pyclpa.util import load_alias, split, join
from pyclpa.base import get_clpa


class Wordlist(list):
    @classmethod
    def from_file(cls, path, sep="\t", comment="#"):
        wl = cls()
        wl.read(path, sep=sep, comment=comment)
        return wl

    def read(self, path, sep="\t", comment="#"):
        with Path(path).open(encoding='utf-8') as handle:
            lines = [unicodedata.normalize('NFC', hline) for hline in handle.readlines()
                     if hline and not hline.startswith(comment)]
        self.extend(list(reader(lines, dicts=True, delimiter=sep)))

    def write(self, path, sep="\t"):
        with UnicodeWriter(path, delimiter=sep) as writer:
            for i, item in enumerate(self):
                if i == 0:
                    writer.writerow(list(item.keys()))
                writer.writerow(list(item.values()))
        if path is None:
            return writer.read()

    def check(self, column='TOKENS', rules=False, clpa=None):
        clpa = clpa or get_clpa()

        if rules:
            rules = load_alias(rules)
            for val in self:
                tokens = [rules[t] if t in rules else t for t in split(val[column])]
                val[column] = join(tokens)

        sounds, errors = {}, Counter({'convertable': 0, 'non-convertable': 0})
        for item in self:
            new_tokens, sounds, errors = clpa.check_sequence(
                split(item[column]), sounds=sounds, errors=errors)
            idxs = [clpa.segment2clpa(t) for t in new_tokens]

            #    new_tokens.append(accent + sounds[token]['clpa'])
            #    idxs.append(sounds[token]['id'])
            item['CLPA_TOKENS'] = join(new_tokens)
            item['CLPA_IDS'] = join(idxs)

        return sounds, errors



# coding: utf8
from __future__ import unicode_literals, print_function, division

from mock import Mock
from clldutils.testing import capture
from clldutils.clilib import ParserError
from clldutils.dsv import reader

from pyclpa.tests.util import TestCase


class Tests(TestCase):
    def _read_tsv(self, path):
        return set(tuple(row[1:]) for row in reader(path, delimiter='\t'))

    def test_report(self):
        from pyclpa.cli import report

        with self.assertRaises(ParserError):
            report(Mock(args=[]))

        with self.assertRaises(ParserError):
            report(Mock(args=['format=csv']))

        args = [self.data_path('KSL.tsv').as_posix()]

        with capture(report, Mock(args=args)) as out:
            self.assertIn('Convertible sounds', out)

        out = self.tmp_path('test.csv')
        report(Mock(args=args + ['format=csv', 'outfile=' + out.as_posix()]))
        self.assertEqual(
            self._read_tsv(out), self._read_tsv(self.data_path('KSL_report.csv')))

        with capture(report, Mock(args=args + ['format=csv'])) as out:
            if hasattr(out, 'decode'):
                out = out.decode('utf8')
            self.assertIn('existing', out)

        with capture(report, Mock(args=args + ['format=cldf'])) as out:
            if hasattr(out, 'decode'):
                out = out.decode('utf8')
            self.assertIn('CLPA_TOKENS', out)

        out = self.tmp_path('test.md')
        report(Mock(args=args + ['outfile=' + out.as_posix()]))
        self.assertTrue(out.exists())

    def test_check(self):
        from pyclpa.cli import check

        with self.assertRaises(ParserError):
            check(Mock(args=[]))

        with capture(check, Mock(args=['abcd'])) as out:
            self.assertIn('?', out)



# coding: utf8
from __future__ import unicode_literals, print_function, division
from collections import Counter

from pyclpa.util import load_alias, load_whitelist, find_token


_clpa = None


def get_clpa():
    global _clpa
    if not _clpa:
        _clpa = CLPA()
    return _clpa


class CLPA(object):
    def __init__(self,
                 whitelist=None,
                 alias=None,
                 delete=None,
                 explicit=None,
                 patterns=None,
                 accents=None,
                 rules=None):
        self.whitelist = whitelist or load_whitelist()
        self.alias = alias or load_alias('alias.tsv')
        self.delete = delete or ['\u0361', '\u035c', '\u0301']
        self.explicit = explicit or load_alias('explicit.tsv')
        self.patterns = patterns or load_alias('patterns.tsv')
        self.accents = accents or "ˈˌ'"
        self.rules = rules or []

    def check_sequence(self, seq, sounds=None, errors=None):
        if not isinstance(seq, (list, tuple)):
            new_seq = seq.split(' ')
        else:
            new_seq = [x for x in seq]
        if self.rules:
            new_seq = [self.rules[t] if t in self.rules else t for t in new_seq]

        new_tokens = []
        sounds = sounds or {}
        errors = errors or Counter({'convertable': 0, 'non-convertable': 0})

        for token in new_seq:
            accent = ''
            if token[0] in self.accents:
                accent, token = token[0], token[1:]

            if token in self.whitelist or token in sounds:
                if token in sounds:
                    sounds[token]['frequency'] += 1
                else:
                    sounds[token] = dict(
                        frequency=1, clpa=token, id=self.whitelist[token]['ID'])
            else:
                check = find_token(
                    token, self.whitelist, self.alias, self.explicit,
                    self.patterns, self.delete)
                sounds[token] = dict(
                    frequency=1,
                    clpa=check if check else '?',
                    id=self.whitelist[check]['ID'] if check else '?')
                errors.update(['convertable' if check else 'non-convertable'])

            new_tokens.append(accent + sounds[token]['clpa'])
        return new_tokens, sounds, errors

    def segment2clpa(self, segment):
        """Convert a segment to it identifier"""
        if segment[0] in self.accents:
            new_segment = segment[1:]
        else:
            new_segment = segment
        return self.whitelist[new_segment]['ID'] \
            if new_segment in self.whitelist else '?'



from setuptools import setup, find_packages


def read(fname):
    with open(fname) as fp:
        return fp.read().split('\n\n\n')[0]

setup(
    name='pyclpa',
    version="0.3.0",
    description='A python library to check phonetic transcriptions',
    long_description=read("README.md"),
    author='Johann-Mattis List',
    author_email='mattis.list@lingpy.org',
    url='https://github.com/glottobank/cpa',
    install_requires=[
        'clldutils',
    ],
    license="GPL",
    zip_safe=False,
    keywords='',
    classifiers=[
        ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'clpa=pyclpa.cli:main',
        ]
    },
    tests_require=['nose', 'coverage', 'mock'],
)
